getPlatform <- function(platform='HM450') {

  require(Biostrings)
  GR <- features(FDb.InfiniumMethylation.hg19) # }}}
  
  if('name' %in% names(values(GR))) { # {{{ for GenomicFeatures < 1.9.18 
    names(GR) <- values(GR)$name
  } # }}}
  if(is.na(genome(GR))) { # {{{ assign genome (GenomicFeatures < 1.9.18) 
    met <- metadata(FDb.InfiniumMethylation.hg19)
    genome(GR) <- met[ which(met[,'name'] == 'Genome'), 'value' ]
  } # }}}
  
  ## this should go into GenomicFeatures in the near future 
  seqinfo(GR) <- SeqinfoForBSGenome( unique(genome(GR)) )[ seqlevels(GR) ]

  platform <- toupper(platform)
  if(platform == 'HM450') { # {{{ assign controls; rename address, channel 
    hm450.controls <- NULL
    data(hm450.controls) 
    GR <- GR[ which(values(GR)$platform %in% c('HM450','BOTH')) ]
    values(GR)$addressA <- Rle(as.numeric(values(GR)$addressA_450))
    values(GR)$addressB <- Rle(as.numeric(values(GR)$addressB_450))
    values(GR)$channel <- Rle(as.factor(values(GR)$channel450))
    attr(GR, 'controls') <- hm450.controls # }}}
  } else if(platform == 'HM27') { # {{{ assign controls; rename address, channel
    hm27.controls <- NULL
    data(hm27.controls)
    GR <- GR[ which(values(GR)$platform %in% c('HM27','BOTH')) ]
    values(GR)$addressA <- Rle(as.numeric(values(GR)$addressA_27))
    values(GR)$addressB <- Rle(as.numeric(values(GR)$addressB_27))
    values(GR)$channel <- Rle(as.factor(values(GR)$channel27))
    attr(GR, 'controls') <- hm27.controls # }}}
  } else { # {{{ complain about not specifying a chip 
    stop('You need to specify either HM27 or HM450 as platform to run')
  } # }}}

  ## this ought to be handled in the database
  values(GR)$percentGC <- as.numeric(values(GR)$percentGC)

  ## this too ought to be handled transparently (and compactly)
  values(GR)$probeType <- Rle(as.factor(values(GR)$probeType))
  values(GR)$platform <- Rle(as.factor(values(GR)$platform))

  ## not so sure whether this will fly in the database though
  values(GR)$sourceSeq <- DNAStringSet(values(GR)$sourceSeq)
  kept = c('addressA','addressB','channel','probeType',
           'platform','percentGC','sourceSeq')
  val <- values(GR)[ , kept ]
  values(GR) <- val
  return(GR)
}

get450k <- function() getPlatform(platform='HM450')

get27k <- function() getPlatform(platform='HM27')

